#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>

#define PAGE_SIZE 4096
#define LWHPS2FPGA_BRIDGE_BASE 0xff200000
#define EXOR_GPIO_OFFSET       0x00010000
#define EXOR_GPIO_IN_OFFSET        0x10
#define EXOR_GPIO_OUT_OFFSET       0x20

volatile unsigned char *gpio_mem;
void *gpio_map;

int main(int argc, char *argv[])
{
	int fd, ret = EXIT_FAILURE;
	unsigned char value;
	unsigned char rvalue;
	off_t gpio_base = LWHPS2FPGA_BRIDGE_BASE + EXOR_GPIO_OFFSET;

	if (argc < 2) {
		fprintf(stderr, "Usage: %s <value>\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	/* check the bounds of the value being set */
	value = atoi(argv[1]);

	/* open the memory device file */
	fprintf(stdout, "Opening /dev/mem file...\n");
	fd = open("/dev/mem", O_RDWR|O_SYNC);
	if (fd < 0) {
		perror("open");
		exit(EXIT_FAILURE);
	}

	/* map the LWHPS2FPGA bridge into process memory */
	fprintf(stdout, "Mapping LWHPS2FPGA bridge into process memory...\n");
	gpio_map = mmap(NULL, PAGE_SIZE, PROT_WRITE, MAP_SHARED,
				fd, gpio_base);
	if (gpio_map == MAP_FAILED) {
		perror("mmap");
		goto cleanup;
	}

	/* get the peripheral's base address */
	fprintf(stdout, "Getting peripheral base address...\n");
	gpio_mem = (unsigned char *)gpio_map;

	/* write the value */
	fprintf(stdout, "Writing value...\n");
	*(gpio_mem + EXOR_GPIO_OUT_OFFSET) = value;

	fprintf(stdout, "Reading back value...\n");
	rvalue = *(gpio_mem + EXOR_GPIO_IN_OFFSET);
	
	fprintf(stdout, "Value written: %d - Readback value: %d\n", value, rvalue);
	
	fprintf(stdout, "Unmapping LWHPS2FPGA bridge...\n");
	if (munmap(gpio_map, PAGE_SIZE) < 0) {
		perror("munmap");
		goto cleanup;
	}

	ret = 0;

cleanup:
	close(fd);
	fprintf(stdout, "Done\n");
	
	return ret;
}
